//#define MOZILLA_STRICT_API

#include "nsIGenericFactory.h"
#include "nsCOMPtr.h"
#include "nsXPCOM.h"
#include "nsIServiceManager.h"
#include "nsICategoryManager.h"
#include "nsIComponentManager.h"
#include "nsIObserver.h"
#include "nsMemory.h"
#include "nsIUSBProfileManager.h"
#include "nsXPCOMCID.h" 
#include "nsISupportsPrimitives.h"
//#include "nsIContentPolicy.h"			// unfrozen component
#include "nsIIOService.h"
//#include "nsIURI.h"
#include "nsString.h"
#include "nsIRegistry.h"
#include "nsXPIDLString.h"
#include "nsIFile.h"
#include "nsILocalFile.h"
#include "nsAppDirectoryServiceDefs.h"
#include "nsDirectoryServiceUtils.h"
//#include "nsIProfileInternal.h"
#include "nsDirectoryServiceDefs.h"
//#include "nsProfileAccess.h"
#include "nsIEnumerator.h"
#include "nsIWindowWatcher.h"
#include "nsIDialogParamBlock.h"
#include "nsIDOMWindowInternal.h"
//#include "nsWindowCreator.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
# include <mntent.h>

static NS_DEFINE_CID(kSupportsCStringCID, NS_SUPPORTS_CSTRING_CID); 

#define USBPROFILE_CID \
{ 0x777f7150, 0x4a2b, 0x4301, \
{ 0xad, 0x10, 0x5e, 0xab, 0x25, 0xb3, 0x22, 0xaa}}

#define USBPROFILE_ContractID "@azhou/USBProfile"

// Registry Keys

#define kRegistryYesString (NS_LITERAL_STRING("yes"))
#define kRegistryNoString (NS_LITERAL_STRING("no"))

#define kRegistryProfileSubtreeString (NS_LITERAL_STRING("Profiles"))
#define kRegistryCurrentProfileString (NS_LITERAL_STRING("CurrentProfile"))
#define kRegistryNCServiceDenialString (NS_LITERAL_STRING("NCServiceDenial"))
#define kRegistryNCProfileNameString (NS_LITERAL_STRING("NCProfileName"))
#define kRegistryNCUserEmailString (NS_LITERAL_STRING("NCEmailAddress"))
#define kRegistryNCHavePREGInfoString (NS_LITERAL_STRING("NCHavePregInfo"))
#define kRegistryHavePREGInfoString (NS_LITERAL_STRING("HavePregInfo"))
#define kRegistryMigratedString (NS_LITERAL_STRING("migrated"))
#define kRegistryDirectoryString (NS_LITERAL_STRING("directory"))
#define kRegistryNeedMigrationString (NS_LITERAL_STRING("NeedMigration"))
#define kRegistryMozRegDataMovedString (NS_LITERAL_STRING("OldRegDataMoved"))
#define kRegistryCreationTimeString (NS_LITERAL_CSTRING("CreationTime"))
#define kRegistryLastModTimeString (NS_LITERAL_CSTRING("LastModTime"))
#define kRegistryMigratedFromString (NS_LITERAL_CSTRING("MigFromDir"))
#define kRegistryVersionString (NS_LITERAL_STRING("Version"))
#define kRegistryVersion_1_0 (NS_LITERAL_STRING("1.0"))
#define kRegistryCurrentVersion (NS_LITERAL_STRING("1.0"))
#define kRegistryStartWithLastString (NS_LITERAL_CSTRING("AutoStartWithLast"))

#define PROFILE_MANAGER_URL "chrome://communicator/content/profile/profileSelection.xul?manage=true"

const char* kDefaultOpenWindowParams = "centerscreen,chrome,modal,titlebar";

/* A mount table entry. */
struct mount_entry
{
  char *me_devname;		/* Device node pathname, including "/dev/". */
  char *me_mountdir;		/* Mount point directory pathname. */
  char *me_type;		/* "nfs", "4.2", etc. */
  dev_t me_dev;			/* Device number of me_mountdir. */
  unsigned int me_dummy : 1;	/* Nonzero for dummy filesystems. */
  unsigned int me_remote : 1;	/* Nonzero for remote fileystems. */
  struct mount_entry *me_next;
};

#ifndef ME_DUMMY
# define ME_DUMMY(fs_name, fs_type) \
    (!strcmp (fs_type, "auto") \
     || !strcmp (fs_type, "autofs") \
     /* for Irix 6.5 */ \
     || !strcmp (fs_type, "ignore"))
#endif

#ifndef ME_REMOTE
# define ME_REMOTE(fs_name, fs_type) (strchr (fs_name, ':') != 0)
#endif

////////////////////// USBProfileManager ////////////////////////////

class nsUSBProfileManager: public nsIObserver, 
		public nsIUSBProfileManager  
{
private:
    PRBool mLocked;

public:
nsUSBProfileManager();
virtual ~nsUSBProfileManager();

NS_DECL_ISUPPORTS
NS_DECL_NSIOBSERVER
NS_DECL_NSIUSBPROFILEMANAGER		

private:
NS_IMETHODIMP SearchForProfile(nsCOMPtr<nsIFile> baseDir);	
PRBool IsProfile(nsCOMPtr<nsIFile> dirEntry);
NS_IMETHODIMP LoadProfile(const PRUnichar *profileName, const PRUnichar *profilePath);
int numOfUSBProfiles;
nsCOMPtr<nsIFile> mNewRegFile;
nsString mUSBProfileName;

// the following fields are for resetting to the last state
nsString prevProfileName;
PRBool prevStartMode;					// record the previous startWithLastUsedProfile flag 
};

nsUSBProfileManager::nsUSBProfileManager()
{
mLocked = PR_TRUE;
numOfUSBProfiles = 0;
NS_GetSpecialDirectory(NS_APP_APPLICATION_REGISTRY_FILE, getter_AddRefs(mNewRegFile));
mUSBProfileName = NS_LITERAL_STRING("");
NS_INIT_ISUPPORTS();
}

nsUSBProfileManager::~nsUSBProfileManager()
{
}

// TODO: need better way to verify. For ex. check whether some required files, dirs exist
PRBool nsUSBProfileManager::IsProfile(nsCOMPtr<nsIFile> dirEntry)
{
	nsCAutoString newPathName, ext;
	nsresult rv = dirEntry->GetNativeLeafName(newPathName);
	if(NS_SUCCEEDED(rv))
	{
		newPathName.Right(ext, 4);
		if(ext.Equals(NS_LITERAL_CSTRING(".slt")))
			return PR_TRUE;
		else
			return PR_FALSE;
	}
	return PR_FALSE;
}

NS_IMETHODIMP nsUSBProfileManager::SearchForProfile(nsCOMPtr<nsIFile> baseDir)
{ 			
     		PRBool hasMore = PR_FALSE;
     		PRBool isDir;
     		nsCOMPtr<nsISimpleEnumerator> dirIterator;
     		nsresult rv = baseDir->GetDirectoryEntries(getter_AddRefs(dirIterator));
     		if (NS_FAILED(rv)) return rv;
     
     		rv = dirIterator->HasMoreElements(&hasMore);
     		if (NS_FAILED(rv)) return rv;
     
     		nsCOMPtr<nsIFile> dirEntry, parentDir; 		
			nsAutoString profileName, profilePath;
//			nsString profileName, profilePath;
    		while (hasMore)
	      { 	
				//printf("more..\n");
				rv = dirIterator->GetNext((nsISupports**)getter_AddRefs(dirEntry));
         	if (NS_SUCCEEDED(rv))
	         {     
	             rv = dirEntry->IsDirectory(&isDir);
	             if (NS_SUCCEEDED(rv))
	             {
	                 if (isDir)
	                 {
	                 		if(IsProfile(dirEntry))
	                 		{	                 			
	         					rv = dirEntry->GetParent(getter_AddRefs(parentDir));
						         if (NS_SUCCEEDED(rv))
						         {
						            numOfUSBProfiles++;
						         	rv = parentDir->GetLeafName(profileName);
						         	rv = dirEntry->GetPath(profilePath);
										
										// load the profile to the registry
						         	rv = LoadProfile(profileName.get(), profilePath.get()); 
						         	
						         	// record the profileName for auto launching
						         	if(NS_SUCCEEDED(rv) && mUSBProfileName.Equals(NS_LITERAL_STRING("")))
											mUSBProfileName = profileName;
						         }                 			
	                 		}
	                 		else
									SearchForProfile(dirEntry);
						  }
					 }
				}
        		rv = dirIterator->HasMoreElements(&hasMore);
	         if (NS_FAILED(rv)) return rv;
	      }	
	      return rv;
}

// Set the directory value and add the entry to the registry tree.
NS_IMETHODIMP nsUSBProfileManager::LoadProfile(const PRUnichar *profileName, const PRUnichar *profilePath)
{
	nsresult rv;
         // TODO: what's this for???
			// append profile name
//         profileDir->Append(nsDependentString(profileName));
//    		rv = profileDir->Exists(&exists);
//         if (NS_FAILED(rv)) return rv;        
//         if (!exists)
//             printf("ProfileDir doesn't exist!\n");		
         
         // TODO: what's this for???
         // since the directory didn't exist, add the indirection		//???
	      //rv = AddLevelOfIndirection(profileDir);
	      //if (NS_FAILED(rv)) return rv;
	               
         // get the profile registry
     		nsCOMPtr<nsIRegistry> registry(do_CreateInstance(NS_REGISTRY_CONTRACTID, &rv));
     		if (NS_FAILED(rv)) return rv;
     		
     		// TODO: need to close the registry???
     		rv = registry->Open(mNewRegFile);
     		if (NS_FAILED(rv)) return rv;

     		nsRegistryKey profilesTreeKey;
     		
     		// Get the major subtree
     		rv = registry->GetKey(nsIRegistry::Common,
                           kRegistryProfileSubtreeString.get(),
                           &profilesTreeKey);
     		if (NS_FAILED(rv))
     		{
         	rv = registry->AddKey(nsIRegistry::Common,
                                 kRegistryProfileSubtreeString.get(),
                                 &profilesTreeKey);
         	if (NS_FAILED(rv)) return rv;
     		}       

			// add the new node to the registry
         nsRegistryKey profKey;
 
         rv = registry->AddKey(profilesTreeKey,			
                                     profileName,
                                     &profKey);
             if (NS_FAILED(rv)) return rv;
 
             rv = registry->SetString(profKey,
                                      kRegistryMigratedString.get(),
												kRegistryYesString.get());	
             if (NS_FAILED(rv)) return rv;
 
             registry->SetString(profKey,
                                 kRegistryNCProfileNameString.get(),
											profileName);
/*	 
             registry->SetString(profKey,
                                 kRegistryNCServiceDenialString.get(),
                                 profileItem->NCDeniedService.get());
 
             registry->SetString(profKey,
                                 kRegistryNCUserEmailString.get(),
                                 profileItem->NCEmailAddress.get());
 
             registry->SetString(profKey,
                                 kRegistryNCHavePREGInfoString.get(),
                                 profileItem->NCHavePregInfo.get());
 
             registry->SetLongLong(profKey,
                                   kRegistryCreationTimeString.get(),
                                   &profileItem->creationTime);
 
             registry->SetLongLong(profKey,
                                   kRegistryLastModTimeString.get(),
                                   &profileItem->lastModTime);
*/ 
				// Externalize the location
 					rv = registry->SetString(profKey,
                                  kRegistryDirectoryString.get(),
                                  profilePath);
                                  
             if (NS_FAILED(rv)) {
                 NS_ASSERTION(PR_FALSE, "Could not update profile location");
                 return rv;
             }
    
// TODO:     profileItem->ExternalizeMigratedFromLocation(registry, profKey);    
	return NS_OK;     
}

NS_IMPL_ISUPPORTS2(nsUSBProfileManager, nsIObserver, nsIUSBProfileManager);
NS_GENERIC_FACTORY_CONSTRUCTOR(nsUSBProfileManager)

///////////////// inplement nsIObserver /////////////////

NS_IMETHODIMP
nsUSBProfileManager::Observe(nsISupports *aSubject, const char *aTopic, const PRUnichar *aData)
{
nsresult rv;
nsCOMPtr<nsIComponentManager> aCompMgr;
NS_GetComponentManager(getter_AddRefs(aCompMgr)); 
if (!aCompMgr)  
    return NS_ERROR_UNEXPECTED;

if(strcmp("http-startup", aTopic) == 0){
printf("**********************************************************\n");
printf("Observe http-startup: Registered...\n");
printf("Topic: %s\n", aTopic);
}

else if(strcmp("app-startup", aTopic) == 0){
printf("**********************************************************\n");
printf("Observe app-startup: Registered...\n");
printf("Topic: %s\n", aTopic);
}

else if(strcmp("xpcom-startup", aTopic) == 0){
printf("**********************************************************\n");
printf("**********************************************************\n");
printf("Observe xpcom-startup: Registered...\n");
printf("Topic: %s\n", aTopic);

	// register for "http-startup" event 
    nsCOMPtr<nsIServiceManager> servman = do_QueryInterface((nsISupports*)aCompMgr, &rv); 
    if (NS_FAILED(rv)) return rv; 
     
    nsCOMPtr<nsICategoryManager> catman; 
    servman->GetServiceByContractID(NS_CATEGORYMANAGER_CONTRACTID,  
                                    NS_GET_IID(nsICategoryManager),  
                                    getter_AddRefs(catman)); 
    if (NS_FAILED(rv)) return rv; 
		char* previous = nsnull;     
	rv = catman->AddCategoryEntry("app-startup", 
                                  "USBProfileManager",  
				  							 USBPROFILE_ContractID,
                                  PR_TRUE,  
                                  PR_TRUE,  
                                  &previous); 
    
    rv = catman->AddCategoryEntry("http-startup-category", 
                                  "USBProfileManager",   
				  							 USBPROFILE_ContractID,
                                  PR_TRUE,  
                                  PR_TRUE,  
                                  &previous); 
    if (previous) 
        nsMemory::Free(previous); 
	 
/* Linked list of mounted filesystems. */
struct mount_entry *me;

/* Read filesystem list. */
  struct mount_entry *mount_list;
  struct mount_entry **mtail = &mount_list;
    struct mntent *mnt;
    char *table = MOUNTED;
    FILE *fp;
    char *devopt;

    fp = setmntent (table, "r");
    if (fp == NULL)
      return NS_OK;					//NS_FAIL?
 	
    while ((mnt = getmntent (fp)))
      {
	me = (struct mount_entry *) malloc (sizeof (struct mount_entry));
	me->me_devname = strdup (mnt->mnt_fsname);
	me->me_mountdir = strdup (mnt->mnt_dir);
	me->me_type = strdup (mnt->mnt_type);
	me->me_dummy = ME_DUMMY (me->me_devname, me->me_type);
	me->me_remote = ME_REMOTE (me->me_devname, me->me_type);
	devopt = strstr (mnt->mnt_opts, "dev=");
	if (devopt)
	  {
	    if (devopt[4] == '0' && (devopt[5] == 'x' || devopt[5] == 'X'))
	      me->me_dev = atoi (devopt + 6);
	    else
	      me->me_dev = atoi (devopt + 4);
	  }
	else
	  me->me_dev = (dev_t) -1;	/* Magic; means not known yet. */

	/* Add to the linked list. */
	*mtail = me;
	mtail = &me->me_next;	
      }
	me = mount_list;

    if (endmntent (fp) == 0)				// What does this mean?
    {
    *mtail = NULL;

    while (mount_list)
      {
	me = mount_list->me_next;
	free (mount_list->me_devname);
	free (mount_list->me_mountdir);
	/* FIXME: me_type is not always malloced.  */
	free (mount_list);
	mount_list = me;
      }
    me = NULL;
  }	
  *mtail = NULL;
 
	/* check if any /dev/sda devices, i.e. USB drives are mounted. */ 
	char* usb_dev = "/dev/sda";
	int len = strlen(usb_dev);
	for (; me; me = me->me_next)
	{
		if(strncmp(me->me_devname, usb_dev, len) == 0)
		{
			printf("%s %s\n", me->me_devname, me->me_mountdir);
 			
 			// For now, only consider one profile situation
 			// look for the .slt folder which is the profile directory
 			nsCOMPtr<nsIFile> baseDir;
			nsAutoString wideString; 
			
			// TODO: change this!!!
			wideString.AssignWithConversion(me->me_mountdir);						
			rv = NS_NewLocalFile(nsDependentString(wideString), PR_TRUE, (nsILocalFile **)((nsIFile **)getter_AddRefs(baseDir)));
			if (NS_FAILED(rv)) return rv;
			
			// search for profiles and load
			SearchForProfile(baseDir);	
		}
	}		
		   printf("Usb Profiles: %d\n", numOfUSBProfiles);
		             
			// get profileName
//       rv = NS_NewLocalFile(nsDependentString(nativeProfileDir), PR_TRUE, (nsILocalFile **)((nsIFile **)getter_AddRefs(profileDir)));
            
/* 
         ProfileStruct* profileItem = new ProfileStruct();
         NS_ENSURE_TRUE(profileItem, NS_ERROR_OUT_OF_MEMORY);
 
     		profileItem->profileName = profileName;
     		profileItem->SetResolvedProfileDir(localFile);
     		profileItem->isMigrated = PR_TRUE;
     		profileItem->isImportType = PR_FALSE;
 
     		// convert "now" from microsecs to millisecs
     		PRInt64 oneThousand = LL_INIT(0, 1000);
     		PRInt64 nowInMilliSecs = PR_Now();
     		LL_DIV(profileItem->creationTime, nowInMilliSecs, oneThousand); 
*/       
         // get the profile registry
     		nsCOMPtr<nsIRegistry> registry(do_CreateInstance(NS_REGISTRY_CONTRACTID, &rv));
     		if (NS_FAILED(rv)) return rv;
     		
     		// TODO: need to close the registry???
     		rv = registry->Open(mNewRegFile);
     		if (NS_FAILED(rv)) return rv;

     		nsRegistryKey profilesTreeKey;
     		
     		// Get the major subtree
     		rv = registry->GetKey(nsIRegistry::Common,
                           kRegistryProfileSubtreeString.get(),
                           &profilesTreeKey);
     		if (NS_FAILED(rv))
     		{
         	rv = registry->AddKey(nsIRegistry::Common,
                                 kRegistryProfileSubtreeString.get(),
                                 &profilesTreeKey);
         	if (NS_FAILED(rv)) return rv;
     		}       			

			// Check if any profile exists
			nsCOMPtr<nsIEnumerator> enumKeys;
    		rv = registry->EnumerateSubtrees( profilesTreeKey, getter_AddRefs(enumKeys));
	      if (NS_FAILED(rv)) return rv;

     		rv = enumKeys->First();
	      if (NS_FAILED(rv)) return rv;
			
			if(NS_OK == enumKeys->IsDone())					// no items
	      {
	      	// Open the profile manager dialog
	      	printf("No profile!!!\n");
  
				// create an nsWindowCreator and give it to the WindowWatcher service
//			   nsIWindowCreator *creatorCallback = new nsIWindowCreator();
//			   if (!creatorCallback)
//			     return NS_ERROR_OUT_OF_MEMORY;
 
//			   nsCOMPtr<nsIWindowCreator> windowCreator(NS_STATIC_CAST(nsIWindowCreator *, creatorCallback));
//			   if (windowCreator) 
//				{	      	
//        			nsCOMPtr<nsIWindowWatcher> windowWatcher(do_GetService(NS_WINDOWWATCHER_CONTRACTID, &rv));		
//        			if(NS_FAILED(rv)) return rv;
//        			windowWatcher->SetWindowCreator(mWindowCreator);
	         
/*	         	nsCOMPtr<nsIDialogParamBlock> ioParamBlock(do_CreateInstance(NS_DIALOGPARAMBLOCK_CONTRACTID, &rv));
	         	if (NS_FAILED(rv)) return rv;
 
	         	// String0  -> mode
	         	// Int0    <-  result code (1 == OK, 0 == Cancel)
 
	         	ioParamBlock->SetNumberStrings(1);
	         	ioParamBlock->SetString(0, NS_LITERAL_STRING("startup").get());
 
	         	nsCOMPtr<nsIDOMWindow> newWindow;
	         	nsCString profileURLStr = NS_LITERAL_CSTRING(PROFILE_MANAGER_URL);
	         
	         
	         	rv = windowWatcher->OpenWindow(nsnull,
	                                        profileURLStr.get(),
	                                        "_blank",
	                                        kDefaultOpenWindowParams,
	                                        ioParamBlock,
	                                        getter_AddRefs(newWindow));
	         	if (NS_FAILED(rv)) return rv;
	         	PRInt32 dialogConfirmed;
	         	ioParamBlock->GetInt(0, &dialogConfirmed);
	         	if (dialogConfirmed == 0) return NS_ERROR_ABORT;	      	
*/					 
					return rv;
//				}
//				else
//					return NS_ERROR_FAILURE;	      
			}
			
			if(numOfUSBProfiles == 1)
			{     		
	      //Get the current profile
         nsXPIDLString tmpCurrentProfile;
			rv = registry->GetString(profilesTreeKey,
                                kRegistryCurrentProfileString.get(),
                                getter_Copies(tmpCurrentProfile));
 
         if (tmpCurrentProfile)  
             prevProfileName = NS_STATIC_CAST(const PRUnichar*, tmpCurrentProfile);   
			    		
	         // Set the current profile
	         rv = registry->SetString(profilesTreeKey,
                                  kRegistryCurrentProfileString.get(),
                                  mUSBProfileName.get());                                  
	         if (NS_FAILED(rv)) return rv;  
            
            // Get the StartWithLastProfile flag
	     		PRInt32 tempLong;
				rv = registry->GetInt(profilesTreeKey,
                           kRegistryStartWithLastString.get(),
                           &tempLong);	            
    			if (NS_SUCCEEDED(rv))
			        prevStartMode = tempLong;  
          
	     		// Set the StartWithLastProfile flag to true only when 
	     		rv = registry->SetInt(profilesTreeKey,
                           kRegistryStartWithLastString.get(),
                           PR_TRUE);	
            if (NS_FAILED(rv)) return rv;  
			}	
			
			// TODO: lock
}

// NEED TO FREE mount_entry *me? //
return NS_OK;
}

static NS_METHOD nsUSBProfileManagerRegistration( 
  nsIComponentManager *aCompMgr, 
  nsIFile *aPath, 
  const char *registryLocation, 
  const char *componentType, 
  const nsModuleComponentInfo *info) 
{ 
    nsresult rv; 
    nsCOMPtr<nsIServiceManager> servman = do_QueryInterface((nsISupports*)aCompMgr, &rv); 
    if (NS_FAILED(rv)) 
        return rv; 
     
    nsCOMPtr<nsICategoryManager> catman; 
    servman->GetServiceByContractID(NS_CATEGORYMANAGER_CONTRACTID,  
                                    NS_GET_IID(nsICategoryManager),  
                                    getter_AddRefs(catman)); 
    if (NS_FAILED(rv)) 
        return rv; 
 
    char* previous = nsnull; 
    rv = catman->AddCategoryEntry("profile-do-change",				
                                  "USBProfileManager",  
				  							 USBPROFILE_ContractID,
                                  PR_TRUE,  
                                  PR_TRUE,  
                                  &previous); 
    if (previous) 
        nsMemory::Free(previous); 

    return rv; 
} 


static NS_METHOD nsUSBProfileManagerUnregistration(nsIComponentManager *aCompMgr,
nsIFile *aPath,
const char *registryLocation,
const nsModuleComponentInfo *info)
{
nsresult rv;

nsCOMPtr<nsIServiceManager> servman = do_QueryInterface((nsISupports*)aCompMgr, &rv);
if (NS_FAILED(rv))
return rv;


nsCOMPtr<nsICategoryManager> catman;
servman->GetServiceByContractID(NS_CATEGORYMANAGER_CONTRACTID,
NS_GET_IID(nsICategoryManager),
getter_AddRefs(catman));

if (NS_FAILED(rv))
return rv;

rv = catman->DeleteCategoryEntry("xpcom-startup",
"USBProfileManager",
PR_TRUE);

rv = catman->DeleteCategoryEntry("content-policy",
"USBProfileManager",
PR_TRUE);

return rv;
}

///////////////// inplement nsIUSBProfileManager /////////////////
// Not implemented

////////////////// nsModuleComponentInfo //////////////////////

static const nsModuleComponentInfo components[] =
{
{ "USBProfile",
USBPROFILE_CID,
USBPROFILE_ContractID,
nsUSBProfileManagerConstructor,
nsUSBProfileManagerRegistration,
nsUSBProfileManagerUnregistration
}
};

NS_IMPL_NSGETMODULE(USBProfileModule, components) 
